;******************************************************************************
;           Bootloader for Microchip PIC16F870/871/873/874/876/877            *           
;                       (c) 2000, Petr Kolomaznik                             *
;                          Freely distributable                               *
;******************************************************************************
;                                                                             *
; Name of file:      bootldr.asm                                              *
; Date:              3.1.2001                                                 *
; Version:           1.03                                                     *
;                                                                             *
; Author:            Petr Kolomaznik                                          *
; Company:           EHL elektronika, Czech Republic                          *
; Email:             kolomaznik@ehl.cz                                        *
; Url:               www.ehl.cz/pic                                           *
;                                                                             *
;******************************************************************************
; How to use this bootloader:                                                 *
; -open project bootldr.pjt in MPLAB                                          *
; -check and modify of parameters in the user setting section with <<< mark   *
; -make project with MPLAB and MPASM                                          *
; -use any programmer for program of bootldr.hex to microcontroller,          *
;  set configuration bits                                                     *
; -use the PIC downloader program for download of user program                *
; -check if user program doesn't use 256 bytes at the top of program memory   *
; -check if goto instruction after RESET (address 0x0000-0x0003) is long jump *
;                                                                             *
; Notes:                                                                      *
; -tab size for editor is 2                                                   *
; -bootloader is compatible with HI-TECH's and Shane Tolmie's bootl./downl.   *
;******************************************************************************
; 1.02 15.11.2000                                                             *
; -check of user constants added                                              *
; -support for the new 16F870/1/2 added                                       *
; -errorlevel directive for message 302 and 306 added                         *
;                                                                             *
; 1.03 3.1.2001                                                               *
; -no support for 16F872 (without USART)                                      *
;	-setting of config bits added                                               *
;                                                                             * 
;******************************************************************************

	errorlevel -302						; no message 302 and 306
	errorlevel -306

	__config _CP_OFF & _DEBUG_OFF & _WRT_ENABLE_ON & _CPD_OFF & _LVP_OFF

;================== User setting section ======================================

	list      p=16f877        ; <<< set type of microcontroller
													  ;     set same microcontroller in the project
	#include <p16f877.inc>    ; <<< set *.inc file for selected microcontroller

	#define FOSC D'3686400'		; <<< set quartz frequence (Hz), max. 20 MHz
	#define BAUD D'9600'			; <<< set baud rate (Bd)
	#define	BAUD_ERROR	D'4'	;	<<< set baud rate error (%) 
	#define TIME							; <<< set method of bootloader start PIN/TIME/AUTO
														;     PIN		start on low level of trigger pin
                            ;     TIME	start on receive IDENT byte in TIMEOUT after RESET
	#define	PORT			PORTB		; <<< only for PIN - set port PORTB/PORTC/PORTD/PORTE
	#define	PORT_BIT	7				; <<< only for PIN - set bit for selected port 0/1/2/3/4/5/6/7
	#define	TIMEOUT		D'3'		; <<< only for TIME - set time (0.1 s), max. 25.5 s

;=============== End of user setting section ==================================

;================== Check User Constants ======================================

	IF FOSC > D'20000000'
		ERROR "max. quartz frequency = 20 MHz"
	ENDIF

	IFNDEF PIN
		IFNDEF TIME
			ERROR "wrong start method of bootloader, must be PIN or TIME"
		ENDIF
	ENDIF

	IF TIMEOUT > D'255'
		ERROR "max. timeout = 25.5 s"
	ENDIF

;========================== Constants =========================================
 
	IF ((FOSC/(D'16' * BAUD))-1) < D'256'
		#define DIVIDER (FOSC/(D'16' * BAUD))-1
		#define HIGH_SPEED 1
	ELSE
		#define DIVIDER (FOSC/(D'64' * BAUD))-1
		#define HIGH_SPEED 0
	ENDIF

#define	BAUD_REAL FOSC/((D'64'-(HIGH_SPEED*D'48'))*(DIVIDER+1))
	IF BAUD_REAL > BAUD
		IF (((BAUD_REAL - BAUD)*D'100')/BAUD) > BAUD_ERROR
			ERROR	"wrong baud rate"
		ENDIF
	ELSE
		IF (((BAUD - BAUD_REAL)*D'100')/BAUD) > BAUD_ERROR
			ERROR	"wrong baud rate"
		ENDIF
	ENDIF		

#define	TIMER	(D'65536'-(FOSC/(D'10'*4*8)))		; reload value for TIMER1 and time 0.1 s

  IFDEF __16F877
    #define LoaderStart 0x1F00
  ELSE
    IFDEF __16F876
      #define LoaderStart 0x1F00
    ELSE
      IFDEF __16F874
        #define LoaderStart 0x0F00
      ELSE
        IFDEF __16F873
          #define LoaderStart 0x0F00
        ELSE
          IFDEF __16F871
            #define LoaderStart 0x0700
          ELSE
            IFDEF __16F870
              #define LoaderStart 0x0700
            ELSE
              ERROR   "incorrect type of microcontroller"
            ENDIF
          ENDIF
        ENDIF
      ENDIF
    ENDIF
  ENDIF

#define LoaderSize	0xFF			  						; size of bootloader	
#define LoaderLOW		LoaderStart+4           ; low address of bootloader
#define LoaderHI		LoaderStart+LoaderSize  ; high address of bootloader

#define READ        0xE0										; communication protocol
#define RACK        0xE1

#define WRITE       0xE3
#define WOK         0xE4
#define WBAD        0xE5

#define DATA_OK     0xE7
#define DATA_BAD    0xE8

#define IDENT       0xEA
#define IDACK       0xEB

#define DONE        0xED

;================ Variables ====================================================

buff				EQU	0x20	     
count				EQU 0x70
amount			EQU	0x71
chk1				EQU	0x72
chk2				EQU	0x73
retry				EQU	0x74
address			EQU 0x75
tmpaddr			EQU 0x77
temp				EQU	0x79
time				EQU	0x7A

;------------------------------------------------------------------------------
		ORG     0x0000                ; reset vector of microcontroller
		pagesel	main	            		
  goto main              					

;------------------------------------------------------------------------------
		ORG			LoaderStart
	goto LoaderStart								; loop for first start without a user program

;------------------------------------------------------------------------------
		ORG     LoaderLOW
main															
	movlw	0x90											; SPEN = 1, CREN = 1
	movwf	RCSTA
	bsf	STATUS,RP0
	IF HIGH_SPEED == 1							; UART SYNC=0; SPEN=1; CREN=1; SREN=0; TXIE=0;
		bsf	TXSTA,BRGH                ;      RCIE=0; TX9=0; RX9=0; TXEN=1;
	ELSE
		bcf TXSTA,BRGH
	ENDIF
	movlw	DIVIDER
	movwf	SPBRG											; baud rate generator
	bcf	TXSTA,SYNC									
	bcf	TXSTA,TX9
	bsf	STATUS,RP0
	bcf	PIE1,TXIE
	bcf	PIE1,RCIE
	bsf	TXSTA,TXEN
	IFDEF PIN												; for PIN
		bsf	PORT,PORT_BIT							; setting of TRIS for selected pin
		bcf	STATUS,RP0								
		btfss	PORT,PORT_BIT
		goto receive									; go to bootloader
		goto	LoaderStart							; run user program	
	ELSE
		bcf	STATUS,RP0								; for TIME
		movlw	TIMEOUT
		movwf	time
		movlw	low TIMER
		movwf	TMR1L
		movlw	high TIMER
		movwf	TMR1H
		movlw	0x31										; prescaler for TIMER1=8 
		movwf	T1CON
		call	getbyte									; wait for IDENT
		xorlw	IDENT
		btfss	STATUS,Z
		goto	LoaderStart
		clrf	time										; no more wait for IDENT
		goto	inst_ident							; bootloader identif., send of IDACK
	ENDIF
																		
;------------------------------------------------------------------------------
receive														; programming	
	call	getbyte										; get of byte from UART
	movwf	temp
	xorlw	WRITE
	btfsc	STATUS,Z
	goto	inst_write								; write instruction
	movf	temp,w
	xorlw	IDENT
	btfsc	STATUS,Z
	goto	inst_ident								; identification instruction
	movf	temp,w
	xorlw	DONE
	btfsc	STATUS,Z
	goto	inst_done									; done instruction
	goto	receive
;------------------------------------------------------------------------------
inst_ident
	movlw	IDACK											; putbyte (IDACK)
	goto	send_byte
;------------------------------------------------------------------------------
inst_write
	call	getbyte										
	movwf	address+1									; high byte of address
	clrf	address
	call	getbyte
	iorwf	address,f									; low byte of address
	call	getbyte										; number of bytes - amount
	movwf	amount
	call	getbyte										; checksum - chk2
	movwf	chk2
	clrf	chk1											; count=chk1=0
	clrf	count
receive_data
	movf	amount,w									; while (count < amount)
	subwf	count,w
	btfsc	STATUS,C
	goto	checksum									; end of receive, check of checksum
	movf	count,w
	addlw	buff
	movwf	FSR
	call	getbyte										; receive of next bytes
	movwf	0
	movf	count,w										; chk1 = chk1 + buff[count++]
	addlw	buff
	movwf	FSR
	movf	0,w
	addwf	chk1,f
	incf	count,f
	goto	receive_data
checksum
	movf	chk1,w										; if (chk1 != chk2)
	xorwf	chk2,w
	btfsc	STATUS,Z
	goto	checksum_ok								; checksum OK
	movlw	DATA_BAD									; checksum WRONG
	goto	send_byte
checksum_ok
	movlw	DATA_OK
	call	putbyte
write_byte
	call	write_eeprom							; write eeprom
	iorlw	0
	btfss	STATUS,Z
	goto	send_wok									; writing OK
	movlw	WBAD											; writing WRONG
	goto	send_byte
;------------------------------------------------------------------------------
inst_done													; end of programming
;------------------------------------------------------------------------------
send_wok
	movlw	WOK												
send_byte
	call	putbyte										; send of byte
	goto	receive										; go to receive from UART
;------------------------------------------------------------------------------

;****************************** putbyte ***************************************
putbyte														; subrutine putbyte
	bcf	STATUS,RP0
	bcf	STATUS,RP1
putbyte1
	clrwdt
	btfss	PIR1,TXIF									; while(!TXIF)
	goto	putbyte1	
	movwf	TXREG											; TXREG = byte
	return

;****************************** getbyte ***************************************
getbyte														; subrutine getbyte
	bcf	STATUS,RP0									
	bcf	STATUS,RP1
getbyte1
	clrwdt
	IFNDEF PIN
		movf	time,w									; for TIME/AUTO
		xorlw	0
		btfsc	STATUS,Z
		goto	getbyte3
		btfss	PIR1,TMR1IF							; check of TIMER1 overflow
		goto	getbyte3								; no overflow
		bcf		T1CON,TMR1ON						; timeout 0.1 s
		decfsz	time,f								; time-1
		goto	getbyte2
		movlw	0
		return
getbyte2
		bcf		PIR1,TMR1IF
		movlw	low TIMER
		movwf	TMR1L
		movlw	high TIMER
		movwf	TMR1H
		bsf		T1CON,TMR1ON						; start TIMER1 
	ENDIF
getbyte3
	btfss	PIR1,RCIF									; while(!RCIF)
	goto	getbyte1
	movf	RCREG,w										; RCREG
	return

;******************************* write eeprom *********************************
write_eeprom											; subrutine write_eeprom
	bcf	STATUS,RP0									
	bcf	STATUS,RP1
	movf	address,w
	movwf	tmpaddr										; tmpaddr = address
	movf	address+1,w
	movwf	tmpaddr+1
	movlw	2													; retry=2
	movwf	retry
	clrf	count											; count=0
w_e_l_1
	movf	amount,w
	subwf	count,w
	btfss	STATUS,C									; while (count<amount)
	goto	w_e_l_2												
	movlw	1													; return 1
	return
w_e_l_2
	movlw	0x21											; if (tmpaddr >= 0x2100) || (tmpaddr <= 0x21FF)
	subwf	tmpaddr+1,w
	btfsc	STATUS,Z
	goto	data_eeprom	
program_eeprom
	bsf	STATUS,RP1
	bsf	STATUS,RP0										
	bsf	EECON1,EEPGD								; EEPGD = 1 - program memory
	bcf	STATUS,RP0
	bcf	STATUS,RP1	
	movlw	high LoaderLOW						; if ((tmpaddr >= LoaderLOW) && (tmpaddr <= LoaderHI))
	subwf	tmpaddr+1,w
	movlw	low LoaderLOW
	btfsc	STATUS,Z
	subwf	tmpaddr,w
	btfss	STATUS,C
	goto	config_word
	movlw	high (LoaderHI+1)
	subwf	tmpaddr+1,w
	movlw	low (LoaderHI+1)
	btfsc	STATUS,Z
	subwf	tmpaddr,w
	btfsc	STATUS,C
	goto	config_word
	goto	next_adr									; next address
config_word												; mask __CONFIG
	movlw	0x20
	subwf	tmpaddr+1,w
	btfsc	STATUS,Z
	goto	next_adr									; (tmpaddr >= 0x2000) && (tmpaddr <= 0x20FF)
	goto	w_e_l_3
data_eeprom
	bsf	STATUS,RP1
	bsf	STATUS,RP0										
	bcf	EECON1,EEPGD								; EEPGD = 0 - data memory
	bcf	STATUS,RP0
	bcf	STATUS,RP1		
w_e_l_3
	movf	tmpaddr,w									
	bsf	STATUS,RP1
	movwf	EEADR											; EEADR = low tmpaddr
	movlw	0													;	if (tmpaddr < 0x0004) EEADRH=0x1F00 >> 8;
	bcf	STATUS,RP1
	subwf	tmpaddr+1,w
	movlw	FSR
	btfsc	STATUS,Z
	subwf	tmpaddr,w
	btfsc	STATUS,C
	goto	w_e_l_4
	movlw	high LoaderStart					; EEADRH = high LoaderStart
	goto	w_e_l_5
w_e_l_4
	movf	tmpaddr+1,w								; EEADRH = high tmpaddr;
w_e_l_5
	bsf	STATUS,RP1
	movwf	EEADRH										; set EEADRH 
	bcf	STATUS,RP1									; EEDATH = buff[count];
	movf	count,w
	addlw	buff
	movwf	FSR
	movf	0,w
	bsf	STATUS,RP1
	movwf	EEDATH										; set EEDATH
	bcf	STATUS,RP1									; EEDATA = buff[count+1];
	incf	count,w
	addlw	buff
	movwf	FSR
	movf	0,w
	bsf	STATUS,RP1
	movwf	EEDATA										; set EEDATA
	bsf	STATUS,RP0										
	bsf	EECON1,WREN									; WREN=1
	movlw	0x55											; EECON2 = 0x55
	movwf	EECON2
	movlw	0xAA											; EECON2 = 0xAA
	movwf	EECON2	
	bsf	EECON1,WR										;	WR=1
	nop															; instructions are ignored	
	nop															; microcontroller waits for a complete write 
	bcf STATUS,RP0
	bcf STATUS,RP1
wait_write
	clrwdt
	btfss	PIR2,EEIF									; necessary for a write to data eeprom 
	goto	wait_write
	bcf	PIR2,EEIF
	bsf STATUS,RP0
	bsf STATUS,RP1		
	bcf	EECON1,WREN									; WREN=0
	bsf	EECON1,RD										; RD=1
	nop	
	nop	
	bcf	STATUS,RP0									; if ((EEDATH != buff[count]) || (EEDATA != buff[count+1]))
	bcf	STATUS,RP1
	movf	count,w
	addlw	buff
	movwf	FSR
	movf	0,w
	bsf	STATUS,RP1
	xorwf	EEDATH,w
	bcf	STATUS,RP1
	btfss	STATUS,Z
	goto	w_e_l_6										; write repeat
	incf	count,w
	addlw	buff
	movwf	FSR
	movf	0,w
	bsf	STATUS,RP1
	xorwf	EEDATA,w
	btfsc	STATUS,Z
	goto	next_adr									; verification OK, next address
	bcf	STATUS,RP1									
w_e_l_6														; write repeat
	decfsz	retry,f
	goto	w_e_l_1
	movlw	0
	return
next_adr
	movlw	2													;	count = count + 2
	bcf	STATUS,RP1
	addwf	count,f
	incf	tmpaddr,f									; tmpaddr = tmpaddr + 1
	btfsc	STATUS,Z
	incf	tmpaddr+1,f
	goto	w_e_l_1

;------------------------------------------------------------------------------
	END

